using Oxide.Core.Libraries.Covalence;
using System.Collections.Generic;
using Newtonsoft.Json;
using Oxide.Core;
using System;

namespace Oxide.Plugins
{
    [Info("ConnectMessages", "Spicy", "1.1.9", ResourceId = 2178)]
    [Description("Provides connect and disconnect messages.")]
    public class ConnectMessages : CovalencePlugin
    {
        #region Country API Class

        private class Response
        {
            [JsonProperty("country")]
            public string Country { get; set; }
        }

        #endregion

        #region Helpers

        private void Broadcast(string key, params object[] args)
        {
            foreach (var player in players.Connected)
                player.Message(string.Format(GetLangValue(key, player.Id), args));
        }

        private bool GetConfigValue(string key) => Config.Get<bool>("Settings", key);

        private string GetLangValue(string key, string userId) => lang.GetMessage(key, this, userId);

        #endregion

        #region Config

        private bool showConnectMessage;
        private bool showConnectCountry;
        private bool showDisconnectMessage;
        private bool showDisconnectReason;
        private bool showAdminMessages;

        protected override void LoadDefaultConfig()
        {
            Config["Settings"] = new Dictionary<string, bool>
            {
                ["ShowConnectMessage"] = true,
                ["ShowConnectCountry"] = true,
                ["ShowDisconnectMessage"] = true,
                ["ShowDisconnectReason"] = true,
                ["ShowAdminMessages"] = true
            };
        }

        private void InitialiseConfig()
        {
            try
            {
                showConnectMessage = GetConfigValue("ShowConnectMessage");
                showConnectCountry = GetConfigValue("ShowConnectCountry");
                showDisconnectMessage = GetConfigValue("ShowDisconnectMessage");
                showDisconnectReason = GetConfigValue("ShowDisconnectReason");
                showAdminMessages = GetConfigValue("ShowAdminMessages");
            }
            catch (InvalidCastException)
            {
                Puts("Your configuration file seems to be invalid. Please make sure only true/false values are being used and check the file isn't corrupt using \"http://pro.jsonlint.com/\".");
            }
        }

        #endregion

        #region Lang

        private new void LoadDefaultMessages()
        {
            lang.RegisterMessages(new Dictionary<string, string>
            {
                ["ConnectMessage"] = "{0} has connected.",
                ["ConnectMessageCountry"] = "<color=#1E90FF>通知：</color><color=#00FFFF>欢迎</color><color=#FFFF00> {0} </color><color=#00FFFF>加入游戏！</color><color=#00FFFF>他来自：{1}</color>",
                ["DisconnectMessage"] = "{0} has disconnected.",
                ["DisconnectMessageReason"] = "<color=#1E90FF>通知：</color><color=#FFFF00> {0} </color><color=#00FFFF>离开了游戏</color><color=#FF0000>{1}</color>"
            }, this, "en");

            lang.RegisterMessages(new Dictionary<string, string>
            {
                ["ConnectMessage"] = "{0} s'est connect�(e).",
                ["ConnectMessageCountry"] = "{0} s'est connect�(e) de {1}.",
                ["DisconnectMessage"] = "{0} s'est disconnect�(e).",
                ["DisconnectMessageReason"] = "{0} s'est disconnect�(e). ({1})"
            }, this);

            lang.RegisterMessages(new Dictionary<string, string>
            {
                ["ConnectMessage"] = "{0} ha conectado.",
                ["ConnectMessageCountry"] = "{0} se ha conectado de {1}.",
                ["DisconnectMessage"] = "{0} se ha desconectado.",
                ["DisconnectMessageReason"] = "{0} se ha desconectado. ({1})"
            }, this);
        }

        #endregion

        #region Hooks

        private void OnServerInitialized()
        {
#if HURTWORLD
            GameManager.Instance.ServerConfig.ChatConnectionMessagesEnabled = false;
#endif
            InitialiseConfig();
        }

        private void OnUserConnected(IPlayer player)
        {
            if (!showConnectMessage || (player.IsAdmin && !showAdminMessages))
                return;

            if (!showConnectCountry)
            {
                Broadcast("ConnectMessage", player.Name.Sanitize());
                return;
            }

            string apiUrl = "http://ip-api.com/json/";

            webrequest.Enqueue(apiUrl + player.Address, null, (code, response) =>
            {
                if (code != 200 || response == null)
                {
                    Puts($"WebRequest to {apiUrl} failed, sending connect message without the country.");
                    Broadcast("ConnectMessage", player.Name.Sanitize());
                    return;
                }

                string country = JsonConvert.DeserializeObject<Response>(response).Country;
                Broadcast("ConnectMessageCountry", player.Name.Sanitize(), country);
            }, this);
        }

        private void OnUserDisconnected(IPlayer player, string reason)
        {
            if (!showDisconnectMessage || (player.IsAdmin && !showAdminMessages))
                return;

            if (!showDisconnectReason)
                Broadcast("DisconnectMessage", player.Name.Sanitize());
            else
                Broadcast("DisconnectMessageReason", player.Name.Sanitize(), reason);
        }

        #endregion
    }
}
