﻿using System;
using System.Collections.Generic;
using System.Linq;
using Newtonsoft.Json;
using Oxide.Core;
using Oxide.Core.Configuration;

// ReSharper disable UnusedMember.Local

namespace Oxide.Plugins
{
    internal class BaseCounter<T>
    {
        public long TimeStampOfJoinServer { get; set; }

        public int TryJoinCounter { get; set; }

        public T Entity { get; set; }

        public BaseCounter(long timeStampOfJoinServer, int tryJoinCounter, T entity)
        {
            TimeStampOfJoinServer = timeStampOfJoinServer;
            TryJoinCounter = tryJoinCounter;
            Entity = entity;
        }
    }

    internal class PlayerInfo
    {
        public string Id { get; set; }

        public string Ip { get; set; }

        public PlayerInfo(string id, string ip)
        {
            Id = id;
            Ip = ip;
        }

        public override string ToString()
        {
            return JsonConvert.SerializeObject(this);
        }
    }

    internal static class BaseCounterEx
    {
        public static BaseCounter<T> TryGet<T>(this List<BaseCounter<T>> counters, T entity)
        {
            return counters.FirstOrDefault(x => x.Entity.Equals(entity));
        }

        public static void EnqueueCounter<T>(this ICollection<BaseCounter<T>> container, T entity)
        {
            BaseCounter<T> playerCounter = new BaseCounter<T>(DateTimeOffset.UtcNow.ToUnixTimeSeconds(), 1, entity);
            container.Add(playerCounter);
        }

        public static void IncreaseJoinTimes<T>(this BaseCounter<T> counter)
        {
            counter.TryJoinCounter++;
        }
    }

	[Info("RustBanner", "Duiwei, Dylech30th", "0.0.1")]
	[Description("kick player if connect too frequent")]
    public class RustBanner : RustPlugin
    {
        private readonly List<BaseCounter<BasePlayer>> _playerList = new List<BaseCounter<BasePlayer>>();
        private readonly List<BaseCounter<string>> _ipList = new List<BaseCounter<string>>();

        private readonly Dictionary<PlayerInfo, int> _kickQueue = new Dictionary<PlayerInfo, int>();
        private readonly DynamicConfigFile _bannedList = Interface.Oxide.DataFileSystem.GetDatafile(SerializeLocation);

        private const string SerializeLocation = "banned-list";
        private const string BanPlayerReason = "you have been permanent banned from server because we detected that your ip or your account connect too frequent, for more information please contact server admin",
            BanCuzPlayerAlreadyExistsReason = "you have been permanent banned from server bacause we detected that you try to connect while you are already in the server, for more information please contact server admin",
            KickPlayerReason = "you have been kicked from server because we detected that your ip or your account connect too frequent",
            KickCuzPlayerAlreadyExistsReason = "you have been kicked because we find that you already had an player entity in server, if you try this action too many times, you will be permanent banned from server";

        ~RustBanner() 
        {
            SerializeBannedList();
        }

        private void SerializeBannedList()
        {
            this._bannedList.Save(SerializeLocation);
        }

        private void OnPlayerInit(BasePlayer player)
        {
			if (CheckIfIpOrSteamIdAlreadyExistsInServer(player))
            {
                KickPlayer(player, KickCuzPlayerAlreadyExistsReason);

                var playerInfo = new PlayerInfo(player.UserIDString, GetPlayerIp(player));

                IncreaseOrInsertKickTimes(playerInfo);
                BanPlayerIfTryTooMuchTimes(playerInfo);
				
				return;
            }
			
            ProcessIp(player);
            ProcessPlayer(player);
        }

        private void ProcessPlayer(BasePlayer player)
        {
            BaseCounter<BasePlayer> playerInfo = this._playerList.TryGet(player);

            if (playerInfo != null)
            {
                playerInfo.IncreaseJoinTimes();
                KickAndBanIfMatchCondition(player);
            }
            else
            {
                this._playerList.EnqueueCounter(player);
            }
        }

        private void ProcessIp(BasePlayer player)
        {
            string ip = GetPlayerIp(player);
            BaseCounter<string> ipInfo = this._ipList.TryGet(ip);

            if (ipInfo != null)
            {
                ipInfo.IncreaseJoinTimes();
                KickAndBanIfMatchCondition(player);
            }
            else
            {
                this._ipList.EnqueueCounter(ip);
            }
        }

        private bool CheckIfIpOrSteamIdAlreadyExistsInServer(BasePlayer player)
        {
            return CheckPlayerExistsBySteamId(player) || CheckPlayerExistsByIp(player);
        }

        private bool CheckPlayerExistsBySteamId(BasePlayer player)
        {
            return this.Player.Players.Count(p => p.UserIDString == player.UserIDString) >= 2;
        }

        private bool CheckPlayerExistsByIp(BasePlayer player)
        {
            return this.Player.Players.Count(p => GetPlayerIp(p) == GetPlayerIp(player)) >= 2;
        }

        private void IncreaseOrInsertKickTimes(PlayerInfo player)
        {
            this._kickQueue[player] = this._kickQueue.ContainsKey(player) ? this._kickQueue[player] + 1 : 1;
        }

        private void BanPlayerIfTryTooMuchTimes(PlayerInfo player)
        {
            const int limit = 5;
            if (this._kickQueue.ContainsKey(player) && this._kickQueue[player] >= limit)
            {
                var p = this.Player.Players.FirstOrDefault(x => GetPlayerIp(x) == player.Ip);
                if (p != null)
                {
                    BanPlayer(p, BanCuzPlayerAlreadyExistsReason);
                }
            }
        }

        private void KickAndBanIfMatchCondition(BasePlayer player)
        {
			if (CounterMatchBanCondition(this._playerList.FirstOrDefault(p => p.Entity == player)) || 
                CounterMatchBanCondition(this._ipList.FirstOrDefault(p => p.Entity == GetPlayerIp(player))))
            {
                KickPlayer(player, KickPlayerReason);
                BanPlayer(player, BanPlayerReason);
            }
        }

        private static bool CounterMatchBanCondition<T>(BaseCounter<T> baseCounter)
        {		
            if (baseCounter == null)
            {
                return false;
            }
            return baseCounter.TryJoinCounter >= 30 && CurrentTimeStamp() - baseCounter.TimeStampOfJoinServer <= 30;
        }

        private void KickPlayer(BasePlayer player, string reason)
        {
            this.Player.Kick(player, reason);
			
			Puts("server kicked player:" + player.displayName);
        }

        private void BanPlayer(BasePlayer player, string reason)
        {
            this.Player.Ban(player, reason);
            this._bannedList[player.UserIDString] = GetPlayerIp(player);
			
			Puts("server banned player:" + player.displayName);
        }

        private string GetPlayerIp(BasePlayer player)
        {
            return this.Player.Address(player);
        }

        private static long CurrentTimeStamp()
        {
            return DateTimeOffset.UtcNow.ToUnixTimeSeconds();
        }
    }
}
